#import required python modules
import numpy as np
import time
import multiprocessing

#import compiled fortran library
import FinProjLib

#import fian project python Library
import FinalProjectLibrary as FPL

#get start time for code timings
tic = time.time()


#perform MD runs for each value of the degree of polymerization
def getRg(M):
    
    #list to store results
    Rgresults = []
    
    #assign bonds
    BondPairs,LJPairs,N,ChainEnds = FPL.AssignBondsDisperse(Nc,M)
    
    #set simulation box size
    L = np.cbrt(N/rho)
    
    #get initial positions
    Pos0 = FPL.InitPositions(N,L)
    #get initial velocities
    Vel = FPL.InitVelocities(N,T)
    
    #relax initial positions to a local minimum
    PEnergy, Pos = FPL.ConjugateGradient(Pos0,dx,EFracTolLS,EFracTolCG,L,Cut,BondPairs,LJPairs)
    #get initial accelerations
    Accel = FPL.InitAccel(Pos,L,Cut,BondPairs,LJPairs)
    
    print "Equilibrating 1/2, M = %.0i, Elapsed Time = %.0f s" % (M,time.time()-tic)
    
    #perform MD equilibration run for the specified number of timesteps
    for i in range(1,NStepsEquil1+1):
        
        #step forward by one block, taking RescaleFreq steps
        Pos,Vel,Accel,KEnergy,PEnergy = FinProjLib.vvintegrate(Pos,Vel,Accel,L,Cut,dt,BondPairs,LJPairs)
        
        #rescale velocities every RescaleFreq steps
        if i % RescaleFreq == 0:    
            Vel = FPL.RescaleVelocities(Vel,T)
        
    #preallocate (NStepsEquil2+1,2) array to store the PEs and KEs for 
    #calculation of the average total energy
    Elist = np.zeros((NStepsEquil2+1,2),dtype=float)
    Elist[0,0] = KEnergy
    Elist[0,1] = PEnergy
        
    print "Equilibrating 2/2, M = %.0i, Elapsed Time = %.0f s" % (M,time.time()-tic)
    
    #perform MD equilibration run for the specified number of timesteps
    for i in range(1,NStepsEquil2+1):
        
        #step forward by one block, taking RescaleFreq steps
        Pos,Vel,Accel,Elist[i,0],Elist[i,1] = FinProjLib.vvintegrate(Pos,Vel,Accel,L,Cut,dt,BondPairs,LJPairs)
        
        #rescale velocities every RescaleFreq steps
        if i % RescaleFreq == 0:    
            Vel = FPL.RescaleVelocities(Vel,T)
        
    #compute mean PEnergy from second equilibration
    EAvg = np.mean(np.sum(Elist,axis=1))
    #compute target kinetic energy
    PEnergy = Elist[-1,1]
    KEtarg = EAvg - PEnergy
    #rescale velocities to net zero momentum
    #Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy and rescale velocities
    KEnergy = 0.5 * np.einsum('ij,ij',Vel,Vel)
    VScale = np.sqrt(KEtarg/KEnergy)
    Vel = Vel * VScale
    
    print "Producing, M = %.0i, Elapsed Time = %.0f s" % (M,time.time()-tic)
    
    #perform MD equilibration run for the specified number of timesteps
    for i in range(1,NStepsProd+1):
        
        #step forward by one block, taking SampleFreq steps
        Pos,Vel,Accel,_,_ = FinProjLib.vvintegrate(Pos,Vel,Accel,L,Cut,dt,BondPairs,LJPairs)
    
        if i % SampleFreq == 0:
            #compute Rg distribution
            Rgresults.append(FPL.RgCalc(Pos,ChainEnds))
    
    print "Done! M = %.0i, Elapsed Time = %.0f s" % (M,time.time()-tic)
    
    #store simulation results in text file
    FPL.ArrayWrite(Rgresults,FileName="Disperse Rg M = "+str(M))
    
    return Rgresults



#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#target number of simulated atoms (may vary due to chain length and dispersity)
Ntarg = 3.
#number density of simulation box
rho = 0.8
#number of simulated chains
Nc = 40
#target system temperature
T = 1.0
#set the frequency of velocity rescaling
RescaleFreq = 50
#line search tolerance
EFracTolLS = 1.e-8
#conjugate gradient tolerance
EFracTolCG = 1.e-10
#line search initial step size
dx = 1.e-3

#========== Get Rg distributions at varying M ==========
#set simulated polymer degrees of polymerization
Mlist = np.repeat(range(2,11),1)
#set time step size
dt = 0.001
#set sample frequency
SampleFreq = 50
#set number of equilibration and production steps
NStepsEquil1 = int(1e4)
NStepsEquil2 = int(1e4)
NStepsProd = int(1e5)
Nblocksprod=int(np.floor(NStepsProd/SampleFreq))+1


if __name__ == '__main__':
        
    print "Starting... Nc = %.0i, Elapsed Time = %.0f s" % (Nc,time.time()-tic)
    
    p = multiprocessing.Pool(8)
    DMresults=p.map_async(getRg,Mlist).get(9e9)
    p.close()
    p.join()
    
#    #visualize difusion coefficients with pyplot
#    #create figure for MSD vs plot
#    DvMplt, DvMax=plt.subplots(figsize=(3.25,3.25))
#    #plot Rg vs M on log-log axes
#    DvMax.plot(Mlist,DMresults,'ko')
#    DvMax.set_xscale('log')
#    DvMax.set_yscale('log')
#    DvMax.yaxis.set_minor_formatter(plt.NullFormatter())
#    DvMax.yaxis.set_major_locator(plt.FixedLocator([2e-2,3e-2]))
#    DvMax.yaxis.set_major_formatter(plt.FixedFormatter(['2', '3']))
#    DvMax.xaxis.set_minor_formatter(plt.NullFormatter())
#    DvMax.xaxis.set_major_locator(plt.FixedLocator([2,3,4,6,10]))
#    DvMax.xaxis.set_major_formatter(plt.FixedFormatter(['2', '3', '4', '6', '$10^{1}$']))
#    #set axes labels
#    plt.xlabel("Degree of Polymerization, $M$",fontsize=10)
#    plt.ylabel("Diffusion Coefficient, $D*10^{2}$",fontsize=10)
#    plt.show()
    
    print "Done! Elapsed Time = %.0f s" % (time.time()-tic)
        
        