#import required python modules
import numpy as np
import time
import os

#import compiled fortran library
import FinProjLib

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

def LineSearch(Pos, Dir, dx, EFracTol, L, Cut, BondPairs, LJPairs,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.einsum('ij,ij',NormDir,NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [FinProjLib.calcenergy(Pos + NormDir * x, L, Cut, BondPairs, LJPairs) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = FinProjLib.calcenergy(Pos + NormDir * dx, L, Cut, BondPairs, LJPairs)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [FinProjLib.calcenergy(Pos + NormDir * 2. * dx, L, Cut, BondPairs, LJPairs)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [FinProjLib.calcenergy(Pos + NormDir * Dists[-1], L, Cut, BondPairs, LJPairs)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = FinProjLib.calcenergy(Pos + NormDir * d3, L, Cut, BondPairs, LJPairs)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = FinProjLib.calcenergy(Pos + NormDir * d3, L, Cut, BondPairs, LJPairs)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG, L, Cut, BondPairs, LJPairs):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    
    #compute the energy and forces at the initial positions
    Forces = np.zeros_like(Pos)
    PE, Forces = FinProjLib.calcenergyforces(Pos, L, Cut, Forces, BondPairs, LJPairs)
    
    #set first line search in the direcion of the initial forces
    #I think there should be a .copy(), but it seems to work better without
    Dir = Forces
    
    #perform iterative line searches until the conjugate gradient tolerance 
    #is reached. Store an arbitrary large OldPE to enter loop.
    OldPE = 1.e300
    while abs((PE - OldPE)/PE) > EFracTolCG:
        
        #store current minimum PE and Forces for comparison later
        OldPE, OldForces = PE, Forces.copy()

        #perform a line serch in the new direction and store the new minimum
        #PE value and its Pos        
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS, L, Cut, BondPairs, LJPairs)
        
        #compute the forces at the new minimum
        _, Forces = FinProjLib.calcenergyforces(Pos, L, Cut, Forces, BondPairs, LJPairs)
        
        #compute the new search direction using the conjugate gradient method
        #np.einsum('ij,ij',OldForces,OldForces) = np.sum(OldForces*OldForces)
        Gamma = np.einsum('ij,ij',(Forces - OldForces),Forces)/np.einsum(
                'ij,ij',OldForces,OldForces)
        Dir = Forces + Gamma*Dir
        
    return PE, Pos

def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    
    #compute integer grid # of locations for cubic lattice
    NLat = int(np.cbrt(N) + 1.)
    LatSpac = L / NLat
    
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    
    #find the total kinetic energy
    KE = 0.5 * np.einsum('ij,ij',Vel,Vel)
    
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    
    return Vel


def InitAccel(Pos,L,Cut,BondPairs,LJPairs):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""

    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = FinProjLib.calcenergyforces(Pos, L, Cut, Forces, BondPairs, LJPairs)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.einsum('ij,ij',Vel,Vel) / (3. * len(Vel))

def AssignBondsPrecise(Nc,M):
    """Returns lists of the bonded and non-bonded pairs of atoms for all pairs 
    in a system of N atoms contained in polymer chains with idential length M.
Input:
    N: Total number of chains in system
    M: Precise polymer chain length
Output:
    BondPairs: (N,_) list of lists giving j bonded to the i given by the index
    LJPairs: (N,_) same as above for j not bonded to i
    N: Total number of atoms in system
"""
    
    #compute number of atoms
    N = Nc * M

    #create lists to store bonded and non-bonded pairs
    BondPairs = [[] for _ in range(N)]
    LJPairs = [[] for _ in range(N)]

    #iterate over all unique pairs of atoms (i,j)
    for i in range(N):
        
        #if atom i is a chain end, then no j>i are bonded
        if (i+1) % M == 0:
            #assign all j>i as non-bonded with i and proceed to next i
            [LJPairs[i].append(j) for j in range(i+1,N)]
            continue
        
        #check every j>i to ensure unique pairs
        for j in range(i+1,N):
                
            #check if atoms j and i are adjacent
            #no need to check for chain end here since that is handled above
            if j == i + 1:
                #assign j == i+1 as a bonded interaction
                BondPairs[i].append(j)
            else:
                #assign (i,j) as non-bonded interaction
                LJPairs[i].append(j)
                
    #pad BondPairs and LJPairs with values of -1 for consistent dimensions
    #max number of j bonded to i for j>i is always less than 1 here
    BPwidth = 1
    LJPwidth = max([len(item) for item in LJPairs])
    BondPairs = np.array([np.pad(item,(0,BPwidth-len(item)),'constant',constant_values=-1) for item in BondPairs],dtype=int)
    LJPairs = np.array([np.pad(item,(0,LJPwidth-len(item)),'constant',constant_values=-1) for item in LJPairs],dtype=int)
    
    return BondPairs, LJPairs, N

def AssignBondsDisperse(Nc,M):
    """Returns lists of the bonded and non-bonded pairs of atoms for all pairs 
    in a system of N atoms contained in polymer chain lengths sampled from a 
    poission distribution with poisson parameter one less than the average chain length.
Input:
    N: Total number of chains in system
    M: Precise polymer chain length
Output:
    BondPairs: (N,_) list of lists giving j bonded to the i given by the index
    LJPairs: (N,_) same as above for j not bonded to i
    N: Total number of atoms in system
    ChainEnds: List of chain end atoms in system
"""
    
    #sample Nc random chain lengths    
    Mlist = np.array([np.max([np.random.poisson(lam=M-1)+1,1]) for _ in range(Nc)],dtype=int)
    
    #compute number of atoms
    N = np.sum(Mlist)

    #create lists to store bonded and non-bonded pairs
    BondPairs = [[] for _ in range(N)]
    LJPairs = [[] for _ in range(N)]
    
    #identify chain ends
    ChainEnds = [np.sum(Mlist[0:i+1])-1 for i,_ in enumerate(Mlist)]

    #iterate over all unique pairs of atoms (i,j)
    for i in range(N):
        
        #if atom i is a chain end, then no j>i are bonded
        if i in ChainEnds:
            #assign all j>i as non-bonded with i and proceed to next i
            [LJPairs[i].append(j) for j in range(i+1,N)]
            continue
        
        #check every j>i to ensure unique pairs
        for j in range(i+1,N):
                
            #check if atoms j and i are adjacent
            #no need to check for chain end here since that is handled above
            if j == i + 1:
                #assign j == i+1 as a bonded interaction
                BondPairs[i].append(j)
            else:
                #assign (i,j) as non-bonded interaction
                LJPairs[i].append(j)
                
    #pad BondPairs and LJPairs with values of -1 for consistent dimensions
    #max number of j bonded to i for j>i is always less than 1 here
    BPwidth = 1
    LJPwidth = max([len(item) for item in LJPairs])
    BondPairs = np.array([np.pad(item,(0,BPwidth-len(item)),'constant',constant_values=-1) for item in BondPairs],dtype=int)
    LJPairs = np.array([np.pad(item,(0,LJPwidth-len(item)),'constant',constant_values=-1) for item in LJPairs],dtype=int)
    
    return BondPairs, LJPairs, N, ChainEnds

def RgCalc(Pos,ChainEnds):
    """Returns the instantaneous temperature.
Input:
    Pos: (N,3) array of atomic positions
    Pos: (Nc,1) array of chain end atoms
Output:
    Rglist: (Nc,1) array of Rg values for each chain in the system
"""
    
    #list to store output Rg    
    Rglist = []
    
    #compute Rg for first chain if neccesary
    if ChainEnds[0] != 0:
        ChainPos = Pos[0:ChainEnds[0]+1]
        Rglist.append(np.sqrt((((ChainPos-ChainPos.mean(axis=0))**2).sum(axis=1)).mean(axis=0)))
        
    #compute Rg for other chains
    for i,ChainEnd in enumerate(ChainEnds[1:]):
        
        #determine starting and stopping atoms for current chain
        start = ChainEnds[i]+1
        stop = ChainEnd
        
        #exclude chains with M = 1
        if start == stop:
            continue
        
        #compute Rg
        ChainPos = Pos[start:stop+1]
        Rglist.append(np.sqrt((((ChainPos-ChainPos.mean(axis=0))**2).sum(axis=1)).mean(axis=0)))
        
    #return Rglist as numpy array
    Rglist = np.array(Rglist, dtype=float)
    
    return Rglist

def ArrayWrite(Array,
               FileName = str(time.time()), Directory = os.getcwd()):
    """Writes a 2-D array to a comma delimited text file. Adjacent columns 
    are separated by commas with each row on a new line. The output file is stored in the working directory
Input:
    Array: (N,M) array to be written to a text file
    FileName: name of the output text file
Output:
    None
"""

    #convert Array to a numpy array
    Array = np.array(Array,dtype=str)
    
    #get length of Array
    N = Array.shape[0]
    
    #open text file to write
    if FileName[-4:] != ".txt":
        FileName = FileName + ".txt"
    txtFile = file(Directory+"\\"+FileName,"w")
    
    #write to text file
    for i in range(N):
        #write current row of array to file as comma separated list
        #include "\n" to move to new line after writting current row
        txtFile.write(",".join(Array[i])+"\n")
    
    #close file
    txtFile.close
    
    return

def ArrayRead(FileName,
              Directory = os.getcwd()):
    """Reads a 2-D array from a comma delimited text file where adjacent 
    columns are separated by commas with each row on a new line. The input 
    file is assumed to be stored in the working directory.
Input:
    FileName: name of the output text file
Output:
    Array: (N,M) array read from the text file
"""    

    
    #Create empty list to store output    
    Array = []
    
    #open text file to write
    if FileName[-4:] != ".txt":
        FileName = FileName + ".txt"
    txtFile = file(Directory+"\\"+FileName,"r")
    
    #write to text file
    for i,row in enumerate(txtFile.readlines()):
        #store current line of txtFile as current row in Array
        Array.append(row.split(","))
    
    #close file
    txtFile.close
    
    #convert Array to a numpy floating point array
    Array = np.array(Array,dtype=float)
    
    return Array

def txtJoin(FileList,
            OutFileName = [], Directory = os.getcwd()):
    """Joins a list of comma delimited text files containing 2-D arrays where 
    adjacent columns are separated by commas with each row on a new line. The 
    input files are assumed to be stored in the working directory.
Input:
    FileList: list of text files to be read
Output:
    None
"""    

    if OutFileName == []:
        OutFileName = "+".join(FileList)
    
    #Create empty list to store output    
    ArrayList = []
    
    #read each text file in list
    [ArrayList.append(ArrayRead(FileName, Directory=Directory)) for FileName in FileList]
    
    #create new concatenated array and write to output text file
    WriteArray = np.vstack(ArrayList)
    ArrayWrite(WriteArray,FileName=OutFileName)
    
    return